/*
    upload64 V1.0, a program to upload Nintendo64 ROM images to the Doctor V64
    This is upload64.c, the main file.
    Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parsys.h"     /* definition of 'system-dependent' functions */

#define max2(a, b) ( (a)>(b) ? (a) : (b) )
#define min2(a, b) ( (a)<(b) ? (a) : (b) )

#define BUFSIZE 32768

#define BYTES2LONG(b, s) ( (((b)[0^(s)] & 0xffL) << 24) | \
                           (((b)[1^(s)] & 0xffL) << 16) | \
                           (((b)[2^(s)] & 0xffL) <<  8) | \
                           (((b)[3^(s)] & 0xffL)) )

#define HEADER_MAGIC 0x80371240

#define SEND_MAX_WAIT 0x10000
#define SYNC_MAX_CNT 8192

void usage(char *progname)
{
  fprintf(stderr, "Usage: %s [-o|-s] [-pPort] [-lLength] <File>\n\n", progname);
  fprintf(stderr, "A program to upload Nintendo64 ROM images to the Doctor V64.\n");
  exit(2);
}

int sendbytes(unsigned int baseport, unsigned int len, char data[], int swap)
{
  long maxwait;
  unsigned int i;

  for( i=0; i<len; i++ ) {
    maxwait = SEND_MAX_WAIT;
    if( (in1byte(baseport+2) & 1) == 0 ) {      /* check ~strobe */
      while( ((in1byte(baseport+2) & 2) != 0) && maxwait-- ) ;  /* wait for */
      if( maxwait <= 0 ) return 1;                        /* auto feed == 0 */
      out1byte(baseport, data[i^swap]);         /* write data    */
      out1byte(baseport+2, 5);                  /* ~strobe = 1   */
    } else {
      while( ((in1byte(baseport+2) & 2) == 0) && maxwait-- ) ;  /* wait for */
      if( maxwait <= 0 ) return 1;                        /* auto feed == 1 */
      out1byte(baseport, data[i^swap]);         /* write data    */
      out1byte(baseport+2, 4);                  /* ~strobe = 0   */
    }
  }
  return 0;
}

int sync_header(unsigned int baseport)
{
  int i=0;

  out1byte(baseport, 0);                        /* data = 00000000    */
  out1byte(baseport+2, 4);                      /* ~strobe=0          */
  while( i<SYNC_MAX_CNT ) {
    if( (in1byte(baseport+2) & 8) == 0 ) {      /* wait for select=0  */
      out1byte(baseport, 0xaa);                 /* data = 10101010    */
      out1byte(baseport+2, 0);                  /* ~strobe=0, ~init=0 */
      while( i<SYNC_MAX_CNT ) {
        if( (in1byte(baseport+2) & 8) != 0 ) {  /* wait for select=1  */
          out1byte(baseport+2, 4);              /* ~strobe=0          */
          while( i<SYNC_MAX_CNT ) {
            if( (in1byte(baseport+2) & 8) == 0 ) { /* w for select=0  */
              out1byte(baseport, 0x55);         /* data = 01010101    */
              out1byte(baseport+2, 0);          /* ~strobe=0, ~init=0 */
              while( i<SYNC_MAX_CNT ) {
                if( (in1byte(baseport+2) & 8) != 0 ) {  /* w select=1 */
                  out1byte(baseport+2, 4);      /* ~strobe=0          */
                  while( i<SYNC_MAX_CNT ) {
                    if( (in1byte(baseport+2) & 8) == 0) { /* select=0 */
                      return 0;
                    }
                    i++;
                  }
                }
                i++;
              }
            }
            i++;
          }
        }
        i++;
      }
      i++;
    }
    i++;
  }
  out1byte(baseport+2, 4);
  return 1;
}

int send_filename(unsigned int baseport, char name[])
{
  int i;
  char *c;
  char mname[12];

  memset(mname, ' ', 11);
  c = max2(strrchr(name, '/'), strrchr(name, '\\'));
  if( c == NULL ) {
    c = name;
  } else {
    c++;
  }
  for( i=0; i<8 && *c!='.' && *c!='\0' ; i++, c++ ) mname[i] = *c;
  c = strrchr(c, '.');
  if( c != NULL ) {
    c++;
    for( i=8; i<11 && *c!='\0'; i++, c++ ) mname[i] = *c;
  }
  return sendbytes(baseport, 11, mname, 0);
}

int send_filelen(unsigned int baseport, long len)
{
  char lenbuffer[4];

  lenbuffer[0] = (len    );
  lenbuffer[1] = (len>> 8);
  lenbuffer[2] = (len>>16);
  lenbuffer[3] = (len>>24);
  return sendbytes(baseport, 4, lenbuffer, 0);
}

int send_header(unsigned int baseport, char name[], long len)
{
  static char protocolId[] = "GD6R\1";

  if( sendbytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( send_filelen(baseport, len) != 0 ) return 1;
  if( send_filename(baseport, name) != 0 ) return 1;
  return 0;
}

int main(int argc, char *argv[])
{
  FILE *file1;
  char *fname1=NULL, *progname=argv[0];
  char buffer1[BUFSIZE];
  long flen1, slen=-1;
  int v64format, swapped=-1;
  unsigned int port=0;

  printf("UPLOAD64 V1.0  Copyright (C) 1997 Andreas Sterbenz (stan@sbox.tu-graz.ac.at)\n");
  printf("This program is released under the terms of the GNU public license.\n\n");

  {
    int i;

    for( i=1; i<argc; i++ ) {
      if( argv[i][0] == '-' ) {
        char *c = argv[i]+1;
        if( *c=='o' || *c=='s' ) {
          if( *(c+1) != '\0' ) usage(progname);
          if( *c=='o' ) {
            swapped = 0;
          } else {
            swapped = 1;
          }
        } else
        if( *c=='p' ) {
          if( (strlen(c)<2) || (strlen(c)>BUFSIZE) ) usage(progname);
          if( (sscanf(c+1, "%x%s", &port, buffer1) != 1) ) usage(progname);
        } else
        if( *c=='l' ) {
          if( *(c+1) == '\0' ) {
            slen = -2;
            continue;
          }
          if( strlen(c)>BUFSIZE ) usage(progname);
          if( sscanf(c+1, "%ld%s", &slen, buffer1) != 1 ) {
            if( sscanf(c+1, "%ldM%s", &slen, buffer1) == 1 ) {
              slen <<= 17;
            } else usage(progname);
          }
        } else usage(progname);
      } else {
        if( fname1 == NULL ) {
          fname1 = argv[i];
        } else {
          usage(progname);
        }
      }
    }
  }
  if( fname1 == NULL ) usage(progname);

  port = getparport(port);
  if( port == 0 ) {
    fprintf(stderr, "%s: Could not detect a parallel port, use commandline to specify one.\n", progname);
    exit(1);
  }
  printf("Using parallel port at %04X.\n", port);

  if( (file1 = fopen(fname1, "rb")) == NULL ) {
    fprintf(stderr, "%s: Could not open file '%s' for reading.\n", progname, fname1);
    exit(1);
  }
  fread(buffer1, 1, 4, file1);
  if( swapped == -1 ) {
    if( BYTES2LONG(buffer1, 0) == HEADER_MAGIC ) {
      swapped = 0;
    } else if( BYTES2LONG(buffer1, 1) == HEADER_MAGIC ) {
      swapped = 1;
    } else {
      fprintf(stderr, "%s: Could not detect type of image '%s', use commandline to override.\n", progname, fname1);
      exit(1);
    }
  }
  v64format = swapped^1;
  printf("The image '%s' is in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
  fseek(file1, 0, SEEK_END);
  flen1 = ftell(file1);
  fseek(file1, 0, SEEK_SET);
  if( slen != -1 ) {
    if( slen == -2 ) {
      slen = (((flen1-1) >>23) +1 ) <<23;
    }
  } else {
    slen = flen1;
  }
  printf("%8ld bytes (%ld.%ld Mbit) will be transmitted.\n", slen, slen>>17, ((slen*100)>>17)%100);

  if( sync_header(port) != 0 ) {
    fprintf(stderr, "%s: Could not establish communication with the Doctor V64!\n", progname);
    fprintf(stderr, "Verify that the V64 is powered on, ready to receive, and that the cable is ok!\n");
    exit(1);
  }
  if( send_header(port, fname1, slen) != 0 ) {
    fprintf(stderr, "%s: Error sending transfer header to the Doctor V64!\n", progname);
    fprintf(stderr, "Please check cables and connection.\n");
    exit(1);
  }
  for( ;; ) {
    unsigned int n;
    static long bytessent=0;

    if( (flen1 > 0) && (slen > 0) ) {
      n = fread(buffer1, 1, min2(flen1, BUFSIZE), file1);
    } else {
      n = min2(slen, BUFSIZE);
    }
    if( n==0 ) {
      if( (flen1>0) && (slen>0) ) printf("\nWARNING: Short read, file may not have been transmitted completely!");
      break;
    }
    if( sendbytes(port, n, buffer1, v64format) != 0 ) {
      fprintf(stderr, "\n%s: Timeout while transmitting data to the Doctor V64!\n", progname);
      fprintf(stderr, "Please check cables and connection.\n");
      exit(1);
    }
    bytessent += n;
    printf("\r%8ld Bytes sent.", bytessent);
    fflush(stdout);
    if( flen1 > 0 ) {
      flen1 -= n;
      if( flen1 <= 0 ) memset(buffer1, 0, BUFSIZE);
    }
    slen -= n;
  }
  printf("\nFile successfully transmitted!\n");
  fclose(file1);
  return 0;
}
